# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Python hafnian wrapper function"""
import unittest

import numpy as np
import hafnian as hf
from hafnian import hafnian, haf_real, haf_complex


class TestVersion(unittest.TestCase):
    """Test the version number is correctly imported"""
    def test_version_number(self):
        """returns true if returns a string"""
        res = hf.version()
        self.assertTrue(isinstance(res, str))


class TestPythonInterfaceWrapper(unittest.TestCase):
    """Various Hafnian consistency checks"""

    def test_array_exception(self):
        """Check exception for non-matrix argument"""
        with self.assertRaises(TypeError):
            hafnian(1)

    def test_square_exception(self):
        """Check exception for non-square argument"""
        A = np.zeros([2, 3])
        with self.assertRaises(ValueError):
            hafnian(A)

    def test_odd_dim_exception(self):
        """Check exception for matrix with odd dimensions"""
        A = np.zeros([3, 3])
        self.assertEqual(hafnian(A), 0)

    def test_non_symmetric_exception(self):
        """Check exception for non-symmetric matrix"""
        A = np.ones([4, 4])
        A[0, 1] = 0.
        with self.assertRaises(ValueError):
            hafnian(A)

    def test_nan(self):
        """Check exception for non-finite matrix"""
        A = np.array([[2, 1], [1, np.nan]])
        with self.assertRaises(ValueError):
            hafnian(A)

    def test_2x2(self):
        """Check 2x2 hafnian"""
        A = np.random.random([2, 2])
        A += A.T
        haf = hafnian(A)
        self.assertEqual(haf, A[0, 1])
        haf = hafnian(A, loop=True)
        self.assertEqual(haf, A[0, 1]+A[0, 0]*A[1, 1])

    def test_3x3(self):
        """Check 2x2 hafnian"""
        A = np.ones([3, 3])
        haf = hafnian(A)
        self.assertEqual(haf, 0.0)
        haf = hafnian(A, loop=True)
        self.assertEqual(haf, 4.0)

    def test_4x4(self):
        """Check 4x4 hafnian"""
        A = np.random.random([4, 4])
        A += A.T
        haf = hafnian(A)
        expected = A[0, 1]*A[2, 3] + \
            A[0, 2]*A[1, 3] + A[0, 3]*A[1, 2]
        self.assertEqual(haf, expected)

        haf = hafnian(A, loop=True)
        expected = A[0, 1]*A[2, 3] \
            + A[0, 2]*A[1, 3] + A[0, 3]*A[1, 2] \
            + A[0, 0]*A[1, 1]*A[2, 3] + A[0, 1]*A[2, 2]*A[3, 3] \
            + A[0, 2]*A[1, 1]*A[3, 3] + A[0, 0]*A[2, 2]*A[1, 3] \
            + A[0, 0]*A[3, 3]*A[1, 2] + A[0, 3]*A[1, 1]*A[2, 2] \
            + A[0, 0]*A[1, 1]*A[2, 2]*A[3, 3]
        self.assertEqual(haf, expected)

    def test_real(self):
        """Check hafnian(A)=haf_real(A) for a random
        real matrix.
        """
        A = np.random.random([6, 6])
        A += A.T
        haf = hafnian(A)
        expected = haf_real(A)
        self.assertEqual(haf, expected)

        haf = hafnian(A, loop=True)
        expected = haf_real(A, loop=True)
        self.assertAlmostEqual(haf, expected, delta=1e-10)

        A = np.random.random([6, 6])
        A += A.T
        A = np.array(A, dtype=np.complex128)
        haf = hafnian(A)
        expected = haf_real(np.float64(A.real))
        self.assertEqual(haf, expected)

    def test_complex(self):
        """Check hafnian(A)=haf_complex(A) for a random
        real matrix.
        """
        A = np.complex128(np.random.random([6, 6]))
        A += 1j*np.random.random([6, 6])
        A += A.T
        haf = hafnian(A)
        expected = haf_complex(A)
        self.assertTrue(np.allclose(haf, expected))

        haf = hafnian(A, loop=True)
        expected = haf_complex(A, loop=True)
        self.assertTrue(np.allclose(haf, expected))


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Hafnian Python interface
========================

This is the top level module of the Hafnian python interface,
containing the function :func:`hafnian`. This function determines,
based on the input matrix, whether to use the complex C hafnian
library, or the slightly faster real C hafnian library.

For more advanced usage, the functions :func:`haf_real` and
:func:`haf_complex` are also provided.

Top-level functions
-------------------

.. autosummary::
   hafnian
   haf_complex
   haf_real
   version

Code details
-------------
"""

import numpy as np

from ._version import __version__
from ._lhafnian import hafnian as haf_complex
from ._rlhafnian import hafnian as haf_real


__all__ = ['hafnian', 'haf_complex', 'haf_real', 'version']


def hafnian(Ain, loop=False, tol=1e-12):
    """Returns the hafnian of matrix A via the C hafnian library.

    .. note::

        If the array is real valued (np.float), the result of
        :func:`haf_real` is returned.

        If the array is complex (np.complex), this function queries
        whether the array Ain has non-zero imaginary part. If so, it
        calls the :func:`haf_complex` function. Otherwise, if all elements
        are exactly real, the :func:`haf_real` function is called.

        For more direct control, you may wish to call :func:`haf_real`
        or :func:`haf_complex` directly.

    Args:
        Ain (array): a square, symmetric array.
        loop (bool): If ``True``, the loop hafnian is returned. Default is ``False``.
        toA (float): the tolerance when checking that the matrix is
            symmetric. Default tolerance is 1e-12.

    Returns:
        np.float64 or np.complex128: the hafnian or loop hafnian of matrix Ain
    """
    # pylint: disable=too-many-return-statements
    if not isinstance(Ain, np.ndarray):
        raise TypeError("Input matrix must be a NumPy array.")

    matshape = Ain.shape

    if matshape[0] != matshape[1]:
        raise ValueError("Input matrix must be square.")

    if np.isnan(Ain).any():
        raise ValueError("Input matrix must not contain NaNs.")

    if np.linalg.norm(Ain-np.transpose(Ain)) >= tol:
        raise ValueError("Input matrix must be symmetric.")

    if matshape[0] % 2 != 0 and not loop:
        return 0.0

    A = Ain

    if matshape[0] %2 != 0 and loop:
        A = np.pad(Ain, pad_width=((0, 1), (0, 1)), mode='constant')
        A[-1, -1] = 1.0

    matshape = A.shape

    if matshape[0] == 2:
        if loop:
            return A[0, 1] + A[0, 0]*A[1, 1]
        return A[0][1]

    if matshape[0] == 4:
        if loop:
            result = A[0, 1]*A[2, 3] \
                + A[0, 2]*A[1, 3] + A[0, 3]*A[1, 2] \
                + A[0, 0]*A[1, 1]*A[2, 3] + A[0, 1]*A[2, 2]*A[3, 3] \
                + A[0, 2]*A[1, 1]*A[3, 3] + A[0, 0]*A[2, 2]*A[1, 3] \
                + A[0, 0]*A[3, 3]*A[1, 2] + A[0, 3]*A[1, 1]*A[2, 2] \
                + A[0, 0]*A[1, 1]*A[2, 2]*A[3, 3]
            return result

        return A[0, 1]*A[2, 3] + A[0, 2]*A[1, 3] + A[0, 3]*A[1, 2]


    if A.dtype == np.complex:
        if np.any(np.iscomplex(A)):
            return haf_complex(A, loop=loop)

        return haf_real(np.float64(A.real), loop=loop)

    return haf_real(A, loop=loop)


def version():
    r"""
    Get version number of Hafnian

    Returns:
      str: The package version number
    """
    return __version__



#!/usr/bin/env python3
import time
from math import factorial

import numpy as np
from scipy import diagonal, randn
from scipy.linalg import qr

import matplotlib.pyplot as plt

from hafnian.lib.librhaf import haf_int, haf_real


a0 = 100.
anm1 = 2.
n = 15
r = (anm1/a0)**(1./(n-1))
nreps = [(int)(a0*(r**((i)))) for i in range(n)]



times = np.empty([n, 2])
for ind, reps in enumerate(nreps):
    size = 2*(ind+1)
    print('\nTesting matrix size {}, with {} reps...'.format(size,reps))

    start = time.time()
    for i in range(reps):
        matrix = np.random.randint(low=-1, high=2, size=[size, size])
        A = np.float64(np.clip(matrix+matrix.T, -1, 1))
        res = haf_real(A)

    end = time.time()
    print('Mean time taken (real): ', (end - start)/reps)
    # print('\t Haf result: ', res)
    times[ind, 0] = (end - start)/reps

    start = time.time()
    for i in range(reps):
        matrix = np.random.randint(low=-1, high=2, size=[size, size])
        A = np.clip(matrix+matrix.T, -1, 1)
        res = haf_int(A)

    end = time.time()
    print('Mean time taken (int): ', (end - start)/reps)
    # print('\t Haf result: ', res)
    times[ind, 1] = (end - start)/reps


plt.semilogy(2*np.arange(1,n+1),times[:, 0], marker='.')
plt.semilogy(2*np.arange(1,n+1),times[:, 1], marker='.')
plt.xlabel(r"Matrix size $n$")
plt.ylabel(r"Time in seconds")
plt.savefig('scipy-timing.png')



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#!/usr/bin/env python3
import sys
import os
from setuptools import setup, Extension
from distutils.command.build_ext import build_ext


class build_ext(build_ext):
    def build_extension(self, ext):
        self._ctypes = isinstance(ext, CTypes)
        return super().build_extension(ext)

    def get_export_symbols(self, ext):
        if self._ctypes:
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        if self._ctypes:
            return ext_name + '.so'
        return super().get_ext_filename(ext_name)


class CTypes(Extension):
    pass


with open("hafnian/_version.py") as f:
	version = f.readlines()[-1].split()[-1].strip("\"'")

# cmdclass = {'build_docs': BuildDoc}

if os.name == 'nt':
    cflags_default = "-std=c99 -O3 -Wall -fPIC -shared -fopenmp -lopenblas"
    libraries = ['openblas']
    extra_link_args = ['-fopenmp', '-lopenblas']
else:
    cflags_default = "-std=c99 -O3 -Wall -fPIC -shared -fopenmp -llapacke"
    libraries = ['lapacke']
    extra_link_args = ['-fopenmp', '-llapacke']

LD_LIBRARY_PATH = os.environ.get('LD_LIBRARY_PATH', "").split(":")
C_INCLUDE_PATH = os.environ.get('C_INCLUDE_PATH', "").split(":")
CFLAGS = os.environ.get('CFLAGS', cflags_default).split()

requirements = [
    "numpy>=1.13"
]

os.environ['OPT'] = ''

info = {
    'name': 'Hafnian',
    'version': version,
    'maintainer': 'Xanadu Inc.',
    'maintainer_email': 'nicolas@xanadu.ai',
    'url': 'http://xanadu.ai',
    'license': 'Apache License 2.0',
    'packages': [
                    'hafnian',
                    'hafnian/tests'
                ],
    # 'package_data': {'hafnian': ['backends/data/*']},
    # 'include_package_data': True,
    'description': 'Open source library for hafnian calculation',
    'long_description': open('README.rst').read(),
    'provides': ["hafnian"],
    'install_requires': requirements,
    # 'extras_require': extra_requirements,
    'ext_package': 'hafnian.lib',
    'ext_modules': [
        CTypes("lhafnian",
            sources=["src/lhafnian.c",],
            depends=["src/lhafnian.h"],
            include_dirs=['/usr/local/include', '/usr/include', './src'] + C_INCLUDE_PATH,
            libraries=libraries,
            library_dirs=['/usr/lib', '/usr/local/lib'] + LD_LIBRARY_PATH,
            extra_compile_args=CFLAGS,
            extra_link_args=extra_link_args
            ),
        CTypes("rlhafnian",
            sources=["src/rlhafnian.c"],
            depends=["src/rlhafnian.h"],
            include_dirs=['/usr/local/include', '/usr/include', './src'] + C_INCLUDE_PATH,
            libraries=libraries,
            library_dirs=['/usr/lib', '/usr/local/lib'] + LD_LIBRARY_PATH,
            extra_compile_args=CFLAGS,
            extra_link_args=extra_link_args
            )
    ],
    'cmdclass': {'build_ext': build_ext},
    'command_options': {
        'build_sphinx': {
            'version': ('setup.py', version),
            'release': ('setup.py', version)}}
}

classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: POSIX",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3 :: Only',
    "Topic :: Scientific/Engineering :: Physics"
]

setup(classifiers=classifiers, **(info))



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Real hafnian function
========================

**Module name:** :mod:`hafnian.rlhaf`

.. currentmodule:: hafnian.rlhaf

Summary
--------

.. autosummary::
    hafnian

Code details
-------------
"""
import ctypes
import os
import numpy as np


path = os.path.dirname(__file__)
sofile = os.path.join(path, "lib/rlhafnian.so")
cdll = ctypes.CDLL(sofile)


_calc_hafnian = cdll.dhaf
_calc_hafnian.restype = ctypes.c_double

_calc_hafnian_loops = cdll.dhaf_loops
_calc_hafnian_loops.restype = ctypes.c_double


def hafnian(l, loop=False):
    """Returns the hafnian of real matrix l via the C hafnian library.

    Args:
        l (array): a real, square, symmetric array of even dimensions.
        loop (bool): If ``True``, the loop hafnian is returned. Default false.

    Returns:
        np.float64: the hafnian of matrix l
    """
    if l.dtype != np.float64:
        l = l.astype(np.float64)
    matshape = l.shape
    a = l.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
    rr = np.float64(np.array([0.0, 0.0]))
    arr = rr.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    if loop:
        res = _calc_hafnian_loops(a, matshape[0], arr) # pylint: disable=unused-variable
    else:
        res = _calc_hafnian(a, matshape[0], arr) # pylint: disable=unused-variable

    return rr[0]



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the lhaf Python function, which calls lhafnian.so"""
import unittest
from math import factorial as fac

import numpy as np
from hafnian._lhafnian import hafnian


hyp1f1 = {
    1: 2.0,
    2: 3.3333333333333335,
    3: 5.0666666666666664,
    4: 7.2761904761904761,
    5: 10.048677248677249,
    6: 13.482635882635883,
    7: 17.689569689569687,
    8: 22.795345888679229,
    9: 28.941685010704621,
    10: 36.287779179502607,
    11: 45.012048825000804,
    12: 55.314048000042611,
    13: 67.416529128084463,
    14: 81.567678700572273,
    15: 98.043536174028574,
    16: 117.1506090894388,
    17: 139.22869825254688,
    18: 164.65394767535315,
    19: 193.84213488876887,
    20: 227.25221819642439
    }


class TestLhaf(unittest.TestCase):
    """Various Hafnian consistency checks"""

    def setUp(self):
        """Set up"""
        self.n = 6

    def test_2x2(self):
        """Check 2x2 hafnian"""
        A = np.complex128(np.random.random([2, 2])) + 1j*np.random.random([2, 2])
        A = A + A.T
        haf = hafnian(A)
        self.assertTrue(np.allclose(haf, A[0, 1]))

    def test_4x4(self):
        """Check 4x4 hafnian"""
        A = np.complex128(np.random.random([4, 4]))
        A += 1j*np.random.random([4, 4])
        A += A.T
        haf = hafnian(A)
        expected = A[0, 1]*A[2, 3] + \
            A[0, 2]*A[1, 3] + A[0, 3]*A[1, 2]
        self.assertTrue(np.allclose(haf, expected))

    def test_identity(self):
        """Check hafnian(I)=0"""
        A = np.identity(self.n)
        haf = hafnian(A)
        self.assertTrue(np.allclose(haf, 0))

    def test_ones(self):
        """Check hafnian(J_2n)=(2n)!/(n!2^n)"""
        A = np.complex128(np.ones([2*self.n, 2*self.n]))
        haf = hafnian(A)
        expected = fac(2*self.n)/(fac(self.n)*(2**self.n))
        self.assertTrue(np.allclose(haf, expected))

    def test_block_ones(self):
        """Check hafnian([[0, I_n], [I_n, 0]])=n!"""
        O = np.zeros([self.n, self.n])
        B = np.ones([self.n, self.n])
        A = np.vstack([np.hstack([O, B]),
                       np.hstack([B, O])])
        A = np.complex128(A)
        haf = hafnian(A)
        expected = float(fac(self.n))
        self.assertTrue(np.allclose(haf, expected))


class TestLhafLoops(unittest.TestCase):
    """Various Hafnian consistency checks"""

    def setUp(self):
        """Set up"""
        self.n = 6

    def test_2x2(self):
        """Check 2x2 loop hafnian"""
        A = np.complex128(np.random.random([2, 2])) + 1j*np.random.random([2, 2])
        A = A + A.T
        haf = hafnian(A, loop=True)
        self.assertTrue(np.allclose(haf, A[0, 1]+A[0, 0]*A[1, 1]))

    def test_4x4(self):
        """Check 4x4 loop hafnian"""
        A = np.complex128(np.random.random([4, 4]))
        A += 1j*np.random.random([4, 4])
        A += A.T
        haf = hafnian(A, loop=True)
        expected = A[0, 1]*A[2, 3] \
            + A[0, 2]*A[1, 3] + A[0, 3]*A[1, 2] \
            + A[0, 0]*A[1, 1]*A[2, 3] + A[0, 1]*A[2, 2]*A[3, 3] \
            + A[0, 2]*A[1, 1]*A[3, 3] + A[0, 0]*A[2, 2]*A[1, 3] \
            + A[0, 0]*A[3, 3]*A[1, 2] + A[0, 3]*A[1, 1]*A[2, 2] \
            + A[0, 0]*A[1, 1]*A[2, 2]*A[3, 3]
        self.assertTrue(np.allclose(haf, expected))

    def test_4x4_zero_diag(self):
        """Check 4x4 loop hafnian with zero diagonals"""
        A = np.complex128(np.random.random([4, 4]))
        A += 1j*np.random.random([4, 4])
        A += A.T
        A -= np.diag(np.diag(A))
        haf = hafnian(A, loop=True)
        expected = A[0, 1]*A[2, 3] + \
            A[0, 2]*A[1, 3] + A[0, 3]*A[1, 2]
        self.assertTrue(np.allclose(haf, expected))

    def test_identity(self):
        """Check loop hafnian(I)=1"""
        A = np.identity(self.n)
        haf = hafnian(A, loop=True)
        self.assertTrue(np.allclose(haf, 1))

    def test_ones(self):
        """Check loop hafnian(J_2n)=hyp1f1(-2n/2,1/2,-1/2)*(2n)!/(n!2^n)"""
        A = np.complex128(np.ones([2*self.n, 2*self.n]))
        haf = hafnian(A, loop=True)
        expected = fac(2*self.n)/(fac(self.n)*(2**self.n))*hyp1f1[self.n]
        self.assertTrue(np.allclose(haf, expected))


if __name__ == '__main__':
    unittest.main()
